//
// CBL_Edition.h
//
// Copyright (c) 2018 Couchbase, Inc All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef COUCHBASE_ENTERPRISE
#define COUCHBASE_ENTERPRISE
#endif

#define CBLITE_VERSION "4.1.0"
#define CBLITE_VERSION_NUMBER 4001000
#define CBLITE_BUILD_NUMBER 11
#define CBLITE_SOURCE_ID "7a4e6e9+176814f"
#define CBLITE_BUILD_TIMESTAMP "2026-06-25T00:26:20Z"
